type sommet = Sommet of string
 and 'a flche == sommet * 'a
 and 'a graphe == (sommet * ('a flche) list) list ;;

(* exemple: K4 *)
(*	let k4 = [	(Sommet "a",[(Sommet "b",true);(Sommet "c",true);(Sommet "d",true)]) ;	*)
(*				(Sommet "b",[(Sommet "c",true);(Sommet "d",true);(Sommet "a",true)]) ;	*)
(*				(Sommet "c",[(Sommet "d",true);(Sommet "a",true);(Sommet "b",true)]) ;	*)
(*				(Sommet "d",[(Sommet "a",true);(Sommet "b",true);(Sommet "c",true)])	] ;;	*)
			
(* exemple: graphe valu et orient *)
(*	let graphe1 = [ (Sommet "a",[(Sommet "f",5)]) ; *)
(*					(Sommet "b",[(Sommet "a",5);(Sommet "f",15)]) ; *)
(*					(Sommet "c",[(Sommet "b",5);(Sommet "f",50)]) ; *)
(*					(Sommet "d",[(Sommet "a",50);(Sommet "b",25);(Sommet "c",20)]) ;	*)
(*					(Sommet "e",[(Sommet "d",10);(Sommet "b",80);(Sommet "f",100)]) ;	*)
(*					(Sommet "f",[]) ] ;;	*)

(*====== matrice_d_incidence_et_sommets ===================================================*)				

(* renvoie un couple (m,s)								*)
(*	o m est la matrice de terme gnrique m(i,j) gal	*)
(*	 l'image par la fonction transf					*)
(*	de la valeur de l'arc (i->j) s'il existe,			*)
(*	ou  nil_du_type sinon								*)
(*	et s est la liste des noms des sommets du graphe	*)
	
let matrice_d_incidence_et_sommets g nil_du_type transf =
	let n = list_length g
	in
	let m = make_matrix n n nil_du_type
	and s = let rec cre_liste_sommets = function
				  [] -> []
				| (Sommet(s),_) :: q -> s :: cre_liste_sommets q
			in cre_liste_sommets g
	in
	let rec ajoute_flche i = function 
		  [] -> ()
		| (Sommet(but),val) :: qf -> 
				m.(i).(index but s) <- transf val ;
				ajoute_flche i qf
	and cre_matrice =	function
			  [] -> ()
			| (Sommet(a),f) :: q -> 
					ajoute_flche (index a s) f ; 
					cre_matrice q
	in
	cre_matrice g ;
	(m,s) ;;

(*====== opration inverse : donner la liste  partir de la matrice =======================*)				

let reprsentation_en_liste m nil_du_type sommets =
	let n = list_length sommets
	and sv = vect_of_list sommets
	in
	let rec cre_liste_de i j petit_accu =
		if j = n then petit_accu
		else cre_liste_de	i (j+1) 
							(	if m.(i).(j) = nil_du_type
								then petit_accu
			 					else ((Sommet(sv.(j)),m.(i).(j)) :: petit_accu)  )
	and cre_liste i accu =
		if i = n then accu
		else cre_liste (i+1) ((Sommet(sv.(i)),(cre_liste_de i 0 [])) :: accu)
	in
	cre_liste 0 [] ;;

(*======== quelques utilitaires pour l'affichage ==========================================*)				
load_object "Utilitaires" ;;
#open "Utilitaires" ;;

(*	let (mk4,_) = matrice_d_incidence_et_sommets k4 false id ;; print_matrice_bool mk4 ;;	*)
(*	let (mg1,_) = matrice_d_incidence_et_sommets graphe1 0 id ;; print_matrice_int mg1 ;;	*)

(*========== quelques fonctions de base trs simples =======================================*)				

(* y a-t-il un arc de a vers b dans le graphe g ? *)
let a_flche_b g a b = try assoc b (assoc a g) ; true with Not_found -> false ;;

(* quel est la valeur attache  l'arc a -> b -- suppos existant -- du graphe g ? *)	
let a_val_b g a b = assoc b (assoc a g) ;;

(* la liste des successeurs/valeurs de a dans g *)
let successeurs_valeurs g a = assoc a g ;;

(* la liste des prdcesseurs/valeurs de a dans g *)
let prdcesseurs_valeurs (g:'a graphe) (a:sommet) =
	let rec prd_rec accu = function
		  [] -> accu
		| (s,ff) :: q ->
			prd_rec	( try (s,(assoc a ff)) :: accu
						  with Not_found -> accu	) 
						q
	in prd_rec [] g ;;

(*	a_flche_b graphe1 (Sommet "a") (Sommet "b") ;; *)
(*	a_val_b graphe1 (Sommet "a") (Sommet "b") ;;	*)
(*	a_flche_b graphe1 (Sommet "b") (Sommet "a") ;; *)
(*	a_val_b graphe1 (Sommet "b") (Sommet "a") ;;	*)
(*	prdcesseurs_valeurs graphe1 (Sommet "a") ;;	*)

(*===========================================================================================*)				
(*																							 *)
(*			recherche des plus courts chemins entre chaque paire de points du graphe		 *)
(*									algorithme de Floyd										 *)
(*																							 *)
(*===========================================================================================*)				

(* on introduit un type correspondant  N U { + } *)
type entier = Infini | N of int ;;

let infrieur x y = match x,y with
	  Infini,Infini -> true
	| Infini,N(_) -> false
	| N(_),Infini -> true
	| N(p),N(q) -> p <= q ;;
	
let strict_infrieur x y = match x,y with
	  Infini,Infini -> false
	| Infini,N(_) -> false
	| N(_),Infini -> true
	| N(p),N(q) -> p < q ;;
	
let plus x y = match x,y with
	  Infini,Infini -> Infini
	| Infini,N(_) -> Infini
	| N(_),Infini -> Infini
	| N(p),N(q) -> N(p + q) ;;

(* pour amliorer la lisibilit *)
type intermdiaire = Direct | Nud of string ;;

(* floyd g renvoie une paire de deux fonctions distance,chemin			 *)
(* distance Sommet(a) Sommet(b) renvoie la plus courte distance de a  b *)
(* chemin Sommet(a) Sommet (b) dcrit le chemin  utiliser				 *)

let floyd g =
	let n = list_length g
	and m,sl = matrice_d_incidence_et_sommets g Infini (function x -> N(x))
	in
	let sv = vect_of_list sl
		(***** a priori les chemins sont tous directs						*)
	and passage = make_matrix n n Direct
	in
		(***** le cur de l'algorithme de Floyd : du O(n^3)  coup sr		*)
	for k = 0 to n-1 do
		for i = 0 to n-1 do
			for j = 0 to n-1 do
				let raccourci = plus m.(i).(k) m.(k).(j)
				in
				if strict_infrieur raccourci m.(i).(j) then
				begin
					m.(i).(j) <- raccourci ;
					passage.(i).(j) <- Nud (sv.(k))
				end
			done
		done
	done ;
	let rec distance (Sommet a) (Sommet b) = m.(index a sl).(index b sl)
		(***** une petite rcursion pour trouver le chemin de a  b			*)
	and chemin (Sommet a) (Sommet b) =
		match passage.(index a sl).(index b sl) with
		Direct -> [(Sommet a) ; (Sommet b)]
		| Nud c -> (chemin (Sommet a)(Sommet c)) @ (tl (chemin (Sommet c)(Sommet b)))
	in
	(distance,chemin) ;;
	
(* exemples *)
(*	print_matrice_int mg1 ;;	*)
(*	let (distance,chemin) = floyd graphe1 ;;	*)
(*	distance (Sommet "a") (Sommet "d") ;;	*)
(*	distance (Sommet "c") (Sommet "f") ;;	*)
(*	chemin (Sommet "c") (Sommet "f") ;; *)

(*===========================================================================================*)				
(*																							 *)
(*					recherche de la fermeture rflexive et transitive du graphe				 *)
(*									algorithme de Warshall									 *)
(*																							 *)
(*===========================================================================================*)				

(* warshall g renvoie une paire de deux fonctions relis,chemin					*)
(* relis est une fonction boolenne qui dit s'il existe un chemin de a vers b	*)
(* chemin Sommet(a) Sommet (b) dcrit le chemin  utiliser						*)

let warshall g =
	let n = list_length g
	and m,sl = matrice_d_incidence_et_sommets g false (function x -> true)
	in
	let sv = vect_of_list sl
	and passage = make_matrix n n Direct
	in
		(***** le cur de l'algorithme de Warshall : du O(n^3)  coup sr	*)
	for k = 0 to n-1 do
		for i = 0 to n-1 do
			for j = 0 to n-1 do
				if m.(i).(k) & m.(k).(j) then
				begin
					m.(i).(j) <- true ;
					passage.(i).(j) <- Nud (sv.(k))
				end
			done
		done
	done ;
	let rec relis (Sommet a) (Sommet b) = m.(index a sl).(index b sl)
		(***** une petite rcursion pour trouver le chemin de a  b			*)
	and chemin (Sommet a) (Sommet b) =
		match passage.(index a sl).(index b sl) with
		Direct -> [(Sommet a) ; (Sommet b)]
		| Nud c -> (chemin (Sommet a)(Sommet c)) @ (tl (chemin (Sommet c)(Sommet b)))
	in
	(relis,chemin) ;;
	
(* exemples *)
(*	print_matrice_int mg1 ;;	*)
(*	let (relis,chemin) = warshall graphe1 ;;	*)
(*	relis (Sommet "a") (Sommet "d") ;; *)
(*	relis (Sommet "c") (Sommet "f") ;; *)
(*	chemin (Sommet "c") (Sommet "f") ;; *)

(*===========================================================================================*)				
(*																							 *)
(*				recherche du chemin le plus court d'un sommet fix aux autres				 *)
(*								algorithme de Dijkstra										 *)
(*																							 *)
(*===========================================================================================*)				

load_object "Files de priorit" ;;
#open "Files de priorit" ;;

(* dijkstra g Sommet(s) renvoie une paire de deux fonctions distance,chemin		*)
(* distance Sommet(but) donne la plus courte distance de Sommet(s)  Sommet(b)	*)
(* chemin Sommet(but) dcrit un chemin qui ralise cette plus courte distance	*)
(* bien sr, l'algorithme de Dijkstra nous interdit les distances ngatives		*)

let dijkstra g (Sommet(nom_source) as source) =
	let n = list_length g
	in
	let distances = make_vect n Infini
	and passage = make_vect n source
	and m,sl = matrice_d_incidence_et_sommets g Infini (function x -> N(x))
	in
	let indice_source = index nom_source sl
	and sv = vect_of_list sl
	(* on fixe le type du tas, et sa taille maximale *)
	and tas = liste_en_tas	( fun (Sommet a) (Sommet b) -> 
								   strict_infrieur distances.(index a sl)
													distances.(index b sl) )
							[source] n
	in
	extrait_minimum tas ; (* on vide le tas *)
	(* initialisation *)
	for i = 0 to n-1 do distances.(i) <- m.(indice_source).(i) done ;
	for i = 0 to n-1 do if i <> indice_source then insre_dans_tas tas (Sommet(sv.(i))) done ;
	for bcl = 1 to n-2 do
		let min = extrait_minimum tas
		in
		match min with Sommet s ->
		let is = index s sl
		and ensemble_C = tas_en_vecteur tas
		in
		for i = 0 to (vect_length ensemble_C) - 1
		do
			let it = index (match ensemble_C.(i) with Sommet(t) -> t) sl
			in
			let somme = plus distances.(is) m.(is).(it)
			in
			if strict_infrieur somme distances.(it) then
			begin
				distances.(it) <- somme ;
				passage.(it) <- Sommet(s) ;
				percolation tas i
			end
		done
	done ;
	let distance = function Sommet(s) -> distances.(index s sl)
	and chemin = function s ->
			let rec cre_chemin (Sommet s) l = 
				let is = index s sl
				in
				if is = indice_source then source :: l
				else cre_chemin (passage.(is)) ((Sommet s) :: l)
			in cre_chemin s []
	in distance,chemin ;;
	
(* exemple: graphe valu et orient *)
(*	let graphe2 = [ (Sommet "a",[(Sommet "b",50);(Sommet "c",30);(Sommet "d",100);(Sommet "e",10)]) ;	*)
(*					(Sommet "b",[]) ;	*)
(*					(Sommet "c",[(Sommet "b",5);(Sommet "d",50)]) ; *)
(*					(Sommet "d",[(Sommet "b",20)]) ;	*)
(*					(Sommet "e",[(Sommet "d",10)]) ] ;; *)
(*	let distance,chemin = dijkstra graphe2 (Sommet "a") ;;	*)

(*===========================================================================================*)				
(*																							 *)
(*				recherche d'un arbre couvrant minimal d'un graphe non orient				 *)
(*								algorithme de Kruskal										 *)
(*																							 *)
(*===========================================================================================*)				

load_object "Partitions" ;;
#open "Partitions" ;;

load_object "Files de priorit" ;;
#open "Files de priorit" ;;

(* kruskal g renvoie un nouveau graphe (en fait un arbre) qui constitue *)
(* un arbre de recouvrement minimal du graphe non orient donn			*)

let kruskal g =
	let rec cre_liste_des_artes accu_final g =
			let rec dvide accu source = function
					  [] -> accu
					| (s,v) :: q -> dvide ((source,s,v)::accu) source q
			in
			match g with 
			  [] -> accu_final
			| (s,l) :: q -> dvide (cre_liste_des_artes accu_final q) s l
	in
	let cre_graphe_depuis_liste l =
			let rec symtrique = function (u,v,w) -> (v,u,w) 
			and gonfle_graphe graphe = function
				  [] -> graphe
				| arc :: q -> gonfle_graphe (ajoute_arete arc (ajoute_arete (symtrique arc) graphe)) q
			and ajoute_arete = function ((((Sommet nom1) as s1),s2,v) as arc) -> function
				  [] -> [ (s1,[(s2,v)]) ]
				| (((Sommet n) as a),l) :: q -> 
					if eq_string n nom1 then (a,(s2,v)::l)::q
					else (a,l)::(ajoute_arete arc q)
			in
			gonfle_graphe [] l
	in
	let rec cre_liste_des_sommets accu = function
		  [] -> accu
		| (s,l) :: q -> cre_liste_des_sommets (s::accu) q
	in
	let la = cre_liste_des_artes [] g 
	and ls = cre_liste_des_sommets [] g
	in
	let tas = liste_en_tas	( fun (_,_,p) (_,_,q) -> p < q ) la (list_length la)
	and les_nuds = liste_en_partition ls (function s -> (index s ls) )
	and ns = list_length ls
	in
	let rec boucle_principale liste n =
		if n < ns - 1 then
		begin
			match (extrait_minimum tas) with (s1,s2,v) ->
			let n1,n2 = (trouve_paquet les_nuds s1),(trouve_paquet les_nuds s2)
			in 
			if n1 <> n2 then
			begin
				fusionne_paquets les_nuds n1 n2 ;
				boucle_principale ((s1,s2,v) :: liste) (n+1)
			end
			else boucle_principale liste n
		end
		else liste
	in
	cre_graphe_depuis_liste (boucle_principale [] 0) ;;

(* exemple *)
(*	let graphe3 =	*)
(*		[	(Sommet "1") , [ (Sommet "2"),1 ; (Sommet "4"),4 ] ;	*)
(*			(Sommet "2") , [ (Sommet "1"),1 ; (Sommet "3"),2 ; (Sommet "4"),6 ; (Sommet "5"),4 ] ;					*)
(*			(Sommet "3") , [ (Sommet "2"),2 ; (Sommet "5"),5 ; (Sommet "6"),6 ] ;									*)
(*			(Sommet "4") , [ (Sommet "1"),4 ; (Sommet "2"),6 ; (Sommet "5"),3 ; (Sommet "7"),4 ] ;					*)
(*			(Sommet "5") , [ (Sommet "2"),4 ; (Sommet "3"),5 ; (Sommet "4"),3 ; (Sommet "6"),8 ; (Sommet "7"),7 ] ; *)
(*			(Sommet "6") , [ (Sommet "3"),6 ; (Sommet "5"),8 ; (Sommet "7"),3 ] ;									*)
(*			(Sommet "7") , [ (Sommet "4"),4 ; (Sommet "5"),7 ; (Sommet "6"),3 ] ] ;;								*)

(*	let min_couvrant3 = kruskal graphe3 ;;	*)

(* on obtiendra le mme rsultat avec l'arbre suivant qui est le mme  l'orientation prs *)
(* Kruskal travaille en effet sur des graphes non orients *)

(*	let graphe4 =	*)
(*		[	(Sommet "1") , [ (Sommet "2"),1 ; (Sommet "4"),4 ] ;													*)
(*			(Sommet "2") , [				  (Sommet "3"),2 ; (Sommet "4"),6 ; (Sommet "5"),4 ] ;					*)
(*			(Sommet "3") , [				  (Sommet "5"),5 ; (Sommet "6"),6 ] ;									*)
(*			(Sommet "4") , [								   (Sommet "5"),3 ; (Sommet "7"),4 ] ;					*)
(*			(Sommet "5") , [													(Sommet "6"),8 ; (Sommet "7"),7 ] ; *)
(*			(Sommet "6") , [								   (Sommet "7"),3 ] ;									*)
(*			(Sommet "7") , [												  ] ] ;;								*)

(*	let min_couvrant4 = kruskal graphe4 ;;	*)

(*===========================================================================================*)				
(*																							 *)
(*				recherche d'un arbre couvrant minimal d'un graphe non orient				 *)
(*								algorithme de Prim											 *)
(*																							 *)
(*===========================================================================================*)				

(* prim g renvoie un nouveau graphe (en fait un arbre) qui constitue *)
(* un arbre de recouvrement minimal du graphe non orient donn		 *)

let prim g =
	let n = list_length g
	in						(* phase d'initialisation *)
	let plus_proche = make_vect n 0
	and dj_choisi = make_vect n false
	and _distance_de = make_vect n (N 0)
	and m,sl = matrice_d_incidence_et_sommets g Infini (function x -> N(x))
	in
	let sv = vect_of_list sl
	in
	for i = 1 to n-1 do _distance_de.(i) <- m.(0).(i) done ;
	dj_choisi.(0) <- true ;
	(* pour construire le graphe rsultat *)
	let cre_graphe_depuis_liste l =
			let rec symtrique = function (u,v) -> (v,u) 
			and gonfle_graphe graphe = function
				  [] -> graphe
				| arc :: q -> gonfle_graphe (ajoute_arete arc (ajoute_arete (symtrique arc) graphe)) q
			and ajoute_arete = fun (i1,i2) g ->
				let s1 = Sommet(sv.(i1)) and s2 = Sommet(sv.(i2)) 
				and v = match m.(i1).(i2) with N(nv) -> nv | Infini -> 999999
				in
				match g with
					  [] -> [ (s1,[(s2,v)]) ]
					| (((Sommet nom),l) as a) :: q -> 
						if i1 = (index nom sl) then (s1,(s2,v)::l)::q
						else a::(ajoute_arete (i1,i2) q)
			in
			gonfle_graphe [] l
	in
	(* on revient  l'algorithme de Prim proprement dit *)
	let rec arte_la_moins_chre i min imin = 
		if i = n then imin
		else	if (not dj_choisi.(i)) & (strict_infrieur _distance_de.(i) min)
				then arte_la_moins_chre (i+1) (_distance_de.(i)) i
				else arte_la_moins_chre (i+1) min imin
	and mise__jour i k =
		if i < n then
		begin
			if (not dj_choisi.(i)) & (strict_infrieur m.(k).(i) _distance_de.(i))
			then ( _distance_de.(i) <- m.(k).(i) ; plus_proche.(i) <- k ) ;
			mise__jour (i+1) k
		end
	in
	let rec boucle_principale liste k =
		if k < n - 1 then
		begin
			let imin = arte_la_moins_chre 1 Infini 0
			in
			dj_choisi.(imin) <- true ;
			mise__jour 1 imin ;
			boucle_principale ((imin,plus_proche.(imin)) :: liste) (k+1)
		end
		else liste
	in
	cre_graphe_depuis_liste (boucle_principale [] 0) ;;

(* exemple *)
(*	let graphe3 =	*)
(*		[	(Sommet "1") , [ (Sommet "2"),1 ; (Sommet "4"),4 ] ;													*)
(*			(Sommet "2") , [ (Sommet "1"),1 ; (Sommet "3"),2 ; (Sommet "4"),6 ; (Sommet "5"),4 ] ;					*)
(*			(Sommet "3") , [ (Sommet "2"),2 ; (Sommet "5"),5 ; (Sommet "6"),6 ] ;									*)
(*			(Sommet "4") , [ (Sommet "1"),4 ; (Sommet "2"),6 ; (Sommet "5"),3 ; (Sommet "7"),4 ] ;					*)
(*			(Sommet "5") , [ (Sommet "2"),4 ; (Sommet "3"),5 ; (Sommet "4"),3 ; (Sommet "6"),8 ; (Sommet "7"),7 ] ; *)
(*			(Sommet "6") , [ (Sommet "3"),6 ; (Sommet "5"),8 ; (Sommet "7"),3 ] ;									*)
(*			(Sommet "7") , [ (Sommet "4"),4 ; (Sommet "5"),7 ; (Sommet "6"),3 ] ] ;;								*)

(*	let min_couvrant3 = prim graphe3 ;; *)
